"""
Orchestration is sorely responsible for integration graph and schedule processing based on the correct order.
This module should ONLY dictate the dependency tree based on processing optimization and not entity relationship.
Orchestration can use higher level datatypes from Model to do garbage collection to optimize for cache datasize.
"""

import sys
import logging
import typing
import enum
import asyncio
import uvloop
import networkx as nx
from dataclasses import dataclass
from multiprocessing import Manager, cpu_count
from concurrent import futures
from genesynth.types import Hashabledict, BaseForeign
from genesynth.graph import Graph, find_node, find_child_node
from genesynth.model import worker, types, BaseDataModel, WorkloadType
from genesynth.extensions import datatypes
from genesynth.worker import Runner
from genesynth.io import load_configs, schema_to_graph
from genesynth.utils import spawn, wait
from genesynth.constraints import *

if sys.version_info < (3, 11):
    uvloop.install()

logger = logging.getLogger(__name__)

class Orchestration:
    """
    Handles processing optimization by determing the type of worker that can be used for each data type.
    """
    def __init__(self, graph, thread=10, runner=Runner(registry=worker)):
        self.graph = graph
        self.runner = runner
        self.max_workers = runner.max_workers * thread
        self.queue = asyncio.Queue(self.max_workers)
        self.executor = futures.ThreadPoolExecutor(thread)

    @classmethod
    def read_dict(cls, data, size=None, name='root'):
        size = size or data['metadata']['size']
        G = nx.DiGraph()
        schema_to_graph(G, name, data, size=int(size), root=name)
        graph = Graph(G, name=name, metadata=data['metadata'])
        #for n in graph.nodes:
        #    if isinstance(n, BaseForeign):
        #        n.resolve_fkey_edge()
        return cls(graph)

    @classmethod
    def read_config(cls, *filenames, name='root'):
        data = load_configs(*filenames)
        size = data['metadata']['size']
        return cls.read_dict(data, size=size)

    async def walk(self):
        # TODO check to make sure this is correct
        for parent, node in self.graph.traversal():
            logger.debug(node)
            await self.queue.put(node)

    @property
    def root(self):
        return next(iter(self.graph.root))

    async def __aiter__(self):
        """
        graph.iter: degree_search -> iterate node -> dependency_graph -> generate
        orchestraion.iter: type check -> aggregate by nearest map/array -> garbage collect
        """
        #for node in self.graph.G:
        #    arr = await self.generate(node)
        #    await node.write(arr)

        # TODO Add reduce step from DataModel and garbage collect
        await self.walk()
        while self.queue.qsize() > 0:
            node = await self.queue.get()
            arr = await self.generate(node)
            logger.debug(arr)
            try:
                await node.write()
            except TypeError:
                await node.write(arr)

    def run(self):
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.get_event_loop()
        if loop and loop.is_running():
            loop.create_task(self.__aiter__())
        else:
            asyncio.run(self.__aiter__())

    async def generate(self, node):
        # TODO replace this with graph traversal and task queue/dequeue
        if isinstance(node, BaseDataModel):
            return await self.asyncio(node)
        elif node.workload == WorkloadType.IO:
            return await self.thread(node)
        elif node.workload == WorkloadType.CPU:
            return await self.process(node)
        else:
            return await self.asyncio(node)

    async def process(self, node):
        return await self.runner.run(node.generate)

    async def thread(self, node):
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.get_event_loop()
        return await loop.run_in_executor(self.executor, spawn, node.generate)

    async def asyncio(self, node):
        return await node.generate()

    def __del__(self, *args):
        del self.graph

def run(config: dict):
    pipe = Orchestration.read_dict(config)
    pipe.run()
    asyncio.run(pipe.root.save())
    return pipe.root



"""
Worker is the uilitiy library that handles how to create worker process or other processing methodologies.
"""

import os
import enum
import asyncio
import functools
from multiprocessing import Manager, cpu_count
from concurrent import futures
from genesynth.utils import spawn

class WorkloadType(enum.Enum):
    DEFAULT = 'asyncio'
    IO = 'thread'
    CPU = 'process'

class WorkerRegistry(dict):
    def register(self, fn):
        self[fn.__qualname__] = fn
        return fn

class Runner:
    def __init__(self, registry, workers=cpu_count()):
        self.registry = registry
        self.max_workers = workers
        self.executor = futures.ProcessPoolExecutor(workers)
        self.methods = {qualname: self._wraps(fn) for qualname, fn in registry.items()}

    @property
    def loop(self):
        return asyncio.get_running_loop()

    def _wraps(self, fn):
        @functools.wraps(fn)
        async def wraps(*args, **kwargs):
            return await asyncio.wrap_future(self.executor.submit(spawn, fn, *args, **kwargs), loop=self.loop)
        return wraps

    async def run(self, method, *args, **kwargs):
        obj = method.__self__ 
        qualname = f'{obj.__class__.__name__}.{method.__name__}'
        if qualname in self.methods:
            return await self.methods[qualname](obj, *args, **kwargs)
        else:
            return await method(*args, **kwargs)



import networkx as nx
from pyvis.network import Network

def render(G, height=600, width=800, relabel=False, **params):
    net = Network(height=height, width=width, directed=True, **params)
    if isinstance(G, str):
        net.from_DOT(G)
        return net
    if relabel:
        #mapping = {n: n.name for n in G}
        mapping = {n: G.nodes._nodes[n]['_id'] for n in G}
        net.from_nx(nx.relabel_nodes(G, mapping))
    else:
        net.from_nx(G)
    return net

def show(G, output='graph.html', relabel=True):
    net = render(G, relabel=relabel, notebook=False)
    net.show(output)



"""
Constraints that governs the relationship between types.
Constraints will govern the following:
Node Level:
    - binds types of different types (swich between type class)
    - default parameter translation
    - masking and sampling
    - ordering
    - *data validation (this will not be provided out of the box)
Edge Level:
    - processing order of operations during orchestration
    - foreign relationship to determine order of operations
"""
import enum
from genesynth.mat import identity, unique, sample, mask, mask_index, null_percent, ordered_index, revert_ordered_index

class Builtin(enum.Enum):
    unique = 1 
    notnull = 2
    nullable = 3
    sorted = 4

class Constraints(enum.Enum):
    ONEOF = 'oneOf'
    SUBSET = 'subset'
    UNIQUE = 'unique'
    NOTNULL = 'notnull'
    UUID = 'uuid'
    REGEXP = 'regexp'
    FORMAT = 'format'
    INCREMENTAL = 'incremental'
    BETWEEN = 'between'

class NodeConstraint:
    pass

class EdgeConstraint:
    pass



import pytest
from pytest import fixture
import asyncio
from genesynth.io import *
from genesynth.graph import *
from genesynth.model import *

@fixture
def basic_graph():
    G = read_dot('tests/graph.dot')
    yield G

@fixture
def G(basic_graph):
    graph = Graph(basic_graph, name='root')
    yield graph 

@fixture(scope='function')
def fixture_graph():
    reseed(1)
    G = nx.DiGraph()
    size = 10
    col1 = SerialFixture(name='col1', size=size)
    col2 = IntegerFixture(name='col2', size=size, min=0, max=10)
    G.add_node(col1)
    G.add_node(col2)
    yield G

@fixture
def model(fixture_graph):
    graph = Graph(fixture_graph, name='table')
    yield graph 

def test_parents(G):
    parent = G.parents('2')
    assert set(parent.nodes) == {'root'}

def test_children(G):
    children = G.children('3')
    assert set(children.nodes) == {'9', '6', '7'}

def test_node_degree(G):
    degree = G.node_degree()
    assert degree['root'] == 0
    assert degree['2'] == 1
    assert degree['9'] == 2
    assert degree['13'] == 3

def test_degree_search(G):
    depth = {degree: nodes for degree, nodes in  G.degree_search()}
    assert depth[0] == ('root',)
    assert '9' in depth[2]
    assert '6' in depth[3]

def test_domain_traversal(G):
    nodes = list(G.domain_traversal('2'))
    assert nodes == [('2', '3'), ('2', '4'), ('2', '5'), ('3', '6'), ('3', '7'), ('7', '9')]

def test_traversal(G):
    nodes = list(G.traversal('2'))
    assert nodes == [('2', '3'), ('3', '6'), ('3', '7'), ('7', '9'), ('2', '4'), ('2', '5')]

def test_generate(model):
    depth = {degree: nodes for degree, nodes in  model.degree_search()}

@fixture
def graph():
    size = 10
    G = nx.DiGraph()
    field1 = BooleanFixture(name='field1', size=size)
    field2 = IntegerFixture(name='field2', size=size, min=0, max=10)
    fields = {'field1': field1, 'field2': field2}
    json = BaseMapFixture(name='json', size=size, children=fields)
    col1 = SerialFixture(name='col1', size=size)
    col2 = IntegerFixture(name='col2', size=size, min=0, max=10)
    children = {'col1': col1, 'col2': col2, 'json': json}
    table = BaseMapFixture(name='table', size=size, children=children)
    G.add_node(table)
    G.add_node(col1)
    G.add_node(col2)
    G.add_edge(table, col1)
    G.add_edge(table, col2)
    G.add_edge(json, field1)
    G.add_edge(json, field2)
    G.add_edge(table, json)
    return G

def test_nested(graph):
    nodes = list(set(graph))
    assert len(nodes) == 6



import pytest
from pytest import fixture
import asyncio
import networkx as nx
from genesynth.types import *

@fixture(scope='function')
def params():
    reseed(1)
    yield {
            'name': 'name',
            'size': 10,
    }

def test_integer(params):
    n = IntegerFixture(min=0, max=100, **params)
    #arr = np.array([37, 12, 72, 9, 75, 5, 79, 64, 16, 1]).astype(str)
    arr = np.array([41, 72, 0, 30, 14, 9, 18, 34, 39, 53]).astype(str)
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_serial(params):
    n = SerialFixture(min=0, step=2, **params)
    arr = np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18]).astype(str)
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_boolean(params):
    n = BooleanFixture(**params)
    arr = np.array(['true', 'true', 'false', 'false', 'true', 'true', 'true', 'true', 'true', 'false'])
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_float(params):
    n = FloatFixture(min=0, max=10, **params)
    #arr = np.array([4.170220, 7.203245, 0.001143748, 3.023326, 1.467559, 0.9233859, 1.862602, 3.455607, 3.967675, 5.388167]).astype(str)
    arr = np.array(['4.17022004702574', '7.203244934421581', '0.0011437481734488664',
                '3.0233257263183977', '1.4675589081711304', '0.923385947687978',
                '1.862602113776709', '3.4556072704304777', '3.9676747423066994',
                '5.3881673400335695'])
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_decimal(params):
    n = DecimalFixture(min=0, max=10, precision=0, scale=5, **params)
    arr = np.array([4.17022, 7.20324, 0.00114, 3.02333, 1.46756, 0.92339, 1.8626, 3.45561, 3.96767, 5.38817]).astype(str)
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_text(params):
    n = BaseTextFixture(length=5, **params)
    arr = np.array(['JXdhV', '5z45g', 'OLTx1', 'DbxAl', 'ecikH', 'AhkoU', 'msHd1', 'spUal', 'kFqgK', '7MvZh'])
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_string(params):
    n = StringFixture(subtype='person', field='first_name', **params)
    arr = np.array(['Walter', 'Hershel', 'Rudolf', 'Shirl', 'Jackeline', 'Ron', 'Martin', 'Afton', 'Keren', 'Zackary'])
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_password(params):
    n = BcryptPassword(**params)
    arr = np.array(['$2b$12$i0VpEBOWfbZAVaBSo63bbH6xnAbnBEoonCrbZINl91huSS6AZPsK2',
                    '$2b$12$0FKcpXzkIRPxBFWGyEbcR8KykF8VH1oF7JCqH7aWY2TYGIAd1JmFE',
                    '$2b$12$wvHMLCbokK1XXYp0PfbaUpgMvejGkqSCtDbxAlg3FmLYbbjSjRQHn',
                    '$2b$12$8XGnOyJtNds82s1t6Uzpa2cY7Jk18RFxvmPAmgPsEu23bmu9WvnPZ',
                    '$2b$12$5v2QE9oSfk4nVL0wvs1L73iIgce1WZvMWxJnfq3I5CrWZaPfh2co9',
                    '$2b$12$AhkoUg4x84spDgOca8sKBtd488gnM8HQPqHtpfr8BON6ytut03suH',
                    '$2b$12$JKpbpeIeeNsXE1jFXe6kW9YtgF5s3i4bt3X40UQlAjSPpd7YIH0Cy',
                    '$2b$12$upbOzJdwihqZyyLoaGFOBQToEDnzI44rOceF2jV2tQ0xRTK137Jkp',
                    '$2b$12$nJUdQSvFkTc8YMq47iW0ORB58xXAkuh47hLzhspUalBbMLZmrHqKp',
                    '$2b$12$QXY8HE1VJxrgYhUH7V8iFJtFwGaBBsyWQEOxmarL2ZF9CZzU9skMG'])
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_timestamp(params):
    n = BaseTimestamp(min='2020-01-01', max='2020-01-10', **params)
    arr = np.array([datetime(2020,1,1), datetime(2020,1,2), datetime(2020,1,3), datetime(2020,1,4), datetime(2020,1,5), datetime(2020,1,6), datetime(2020,1,7), datetime(2020,1,8), datetime(2020,1,9), datetime(2020,1,10)]).astype(str)
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_date(params):
    n = BaseDate(min='2020-01-01', max='2020-01-10', **params)
    arr = np.array([date(2020,1,1), date(2020,1,2), date(2020,1,3), date(2020,1,4), date(2020,1,5), date(2020,1,6), date(2020,1,7), date(2020,1,8), date(2020,1,9), date(2020,1,10)]).astype(str)
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_time(params):
    n = BaseTime(min=time(0), max=time(23, 59, 59), **params)
    arr = np.array([time(0), time(2, 39, 59, 888888), time(5, 19, 59, 777777), time(7, 59, 59, 666666), time(10, 39, 59, 555555), time(13, 19, 59, 444444), time(15, 59, 59, 333333), time(18, 39, 59, 222222), time(21, 19, 59, 111111), time(23, 59, 59)]).astype(str)
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_foreign(params):
    parent = IntegerFixture(min=0, max=100, name='parent', size=params['size'])
    child = IntegerFixture(min=0, max=100, name='parent.child', size=params['size'])
    G = nx.DiGraph()
    G.add_edge(parent, child)
    n = BaseForeign.from_params(graph=G, depends_on='parent.child', **params)
    #arr = np.array([37, 12, 72, 9, 75, 5, 79, 64, 16, 1]).astype(str)
    arr = np.array([41, 72, 0, 30, 14, 9, 18, 34, 39, 53]).astype(str)
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_map(params):
    children = {
        'id': SerialFixture(name='id', size=params['size'], min=0, step=1),
        'value': IntegerFixture(name='value', size=params['size'], min=0, max=10),
    }
    n = BaseMapFixture(children=children, **params)
    records = asyncio.run(n.generate())
    assert len(records) == len(children)
    assert set(records) == {'id', 'value'}
    np.testing.assert_array_equal(records['id'], np.arange(10).astype(str))

def test_array(params):
    children = [
        SerialFixture(name='id', size=params['size'], min=0, step=1),
        IntegerFixture(name='value', size=params['size'], min=0, max=10),
    ]
    n = BaseArrayFixture(children=children, **params)
    records = asyncio.run(n.generate())
    assert len(records) == len(children)
    np.testing.assert_array_equal(records[0], np.arange(10).astype(str))

def test_nest(params):
    fields = {
        'value': IntegerFixture(name='value', size=params['size'], min=0, max=10),
    }
    children = {
        'id': SerialFixture(name='id', size=params['size'], min=0, step=1),
        'json': BaseMapFixture(name='json', size=params['size'], children=fields),
    }
    n = BaseMapFixture(children=children, **params)
    records = asyncio.run(n.generate())
    assert set(records) == {'id', 'json'}
    np.testing.assert_array_equal(records['id'], np.arange(10).astype(str))
    assert set(records['json']) == {'value'}
    #np.testing.assert_array_equal(records['json']['value'], np.array([5, 8, 9, 5, 0, 0, 1, 7, 6, 9]).astype(str))
    np.testing.assert_array_equal(records['json']['value'], np.array([4, 6, 2, 8, 0, 6, 4, 5, 1, 1]).astype(str))



import re
from itertools import groupby
from functools import wraps
from contextlib import contextmanager
import asyncio
from concurrent import futures

class classproperty:
    def __init__(self, fn):
        self.fn = fn
    def __get__(self, obj, cls):
        return self.fn(cls)

class Hashabledict(dict):
    def __hash__(self):
        return hash(frozenset(self))

def clean_text(line):
    line = re.sub(r'"(?P<num>[-0-9.]+)"', r'\g<num>', line)
    line = re.sub(r'"(?P<bool>true|false)"', r'\g<bool>', line)
    return line

def sorted_groupby(arr, func, reverse=False):
    return groupby(sorted(arr, key=func, reverse=reverse), key=func)

def spawn(fn, *args, **kwargs):
    return asyncio.run(fn(*args, **kwargs))

async def waits(*futures, timeout=None):
    done, notdone = futures.wait(futures, timeout=timeout, return_when='FIRST_COMPLETED')
    return done.pop().result()

async def wait(future, check_interval=0.1):
    while not future.done():
        await asyncio.sleep(check_interval)
    return future.result()

def iterate_lines(*filenames):
    filehandlers = []
    for filename in  filenames:
        filehandlers.append(open(filename, 'rb'))
    for lines in zip(*filehandlers):
        yield [line.rstrip(b'\n') for line in lines]




#!/usr/bin/env python
import sys
import os
import json
import logging
import argparse
import uvicorn
from fastapi import FastAPI
from pydantic import BaseModel
from genesynth.orchestration import *
from genesynth.utils import clean_text

parser = argparse.ArgumentParser()
parser.add_argument('--host', default='localhost', help='server host')
parser.add_argument('-p', '--port', default=8080, type=int, help='server port')
parser.add_argument('-l', '--level', default='INFO', help='log level')

logger = logging.getLogger(__name__)
app = FastAPI()

class Schema(BaseModel):
    type: str = 'json'
    metadata: dict = {}
    constraints: list = []
    properties: dict = {}

@app.get('/')
async def home():
    return {'status': 'OK'}

@app.post('/api')
async def api(schema: Schema, size=None, string=False):
    pipe = Orchestration.read_dict(dict(schema), size=size)
    pipe.run()
    await pipe.root.save() 
    with open(pipe.root._file) as fh:
        try:
            if string:
                data = [json.loads(line) for line in fh if line.strip()]
            else:
                data = [json.loads(clean_text(line)) for line in fh if line.strip()]
        except:
            fh.seek(0)
            if string:
                data = [line for line in fh if line.strip()]
            else:
                data = [clean_text(line) for line in fh if line.strip()]
        del pipe
        return data

def run_server(host='localhost', port=8080, level='INFO'):
    debug = level == 'DEBUG'
    logging.basicConfig(level=level)
    uvicorn.run('__main__:app', host=host, port=port, log_level=level.lower(), reload=debug)

if __name__ == '__main__':
    args = parser.parse_args()
    run_server(host=args.host, port=args.port, level=args.level)



import pytest
from pytest import fixture
import numpy as np
from genesynth.types import reseed
from genesynth.mat import *

@fixture(scope='function')
def arr():
    reseed(1)
    #arr = np.random.randint(0, 10, 10) # array([5, 8, 9, 5, 0, 0, 1, 7, 6, 9])
    arr = np.array([5, 8, 9, 5, 0, 0, 1, 7, 6, 9])
    yield arr

def test_mask(arr):
    value = mask(arr, 9)
    np.testing.assert_array_almost_equal(value, [2, 9])

def test_null_percent(arr):
    mask = null_percent(arr.size, 50)
    assert mask.sum() == 2
    np.testing.assert_array_equal(mask.data, [False,  True, False, False, False, False, False, False, False, True])

def test_indexed(arr):
    idx = np.random.randint(0, 10, 10)
    array = indexed(arr, idx)
    np.testing.assert_array_equal(array, [0, 6, 9, 0, 5, 5, 8, 7, 1, 9])

def test_ordered_index(arr):
    idx = ordered_index(arr)
    np.testing.assert_array_equal(idx, [4, 5, 6, 0, 3, 8, 7, 1, 2, 9])

def test_revert_ordered_index(arr):
    idx = revert_ordered_index(arr)
    np.testing.assert_array_equal(idx, [3, 7, 8, 4, 0, 1, 2, 6, 5, 9])



import pytest
from pytest import fixture
import tempfile
import yaml
import asyncio
import networkx as nx
import numpy as np
from genesynth.io import *
from genesynth.orchestration import *

@fixture
def config(): 
    yield load_config('tests/test.yaml')

@fixture
def table(config):
    for table, params in config['properties'].items():
        break
    return params

@fixture
def file():
    with open('foo', 'w+') as fh:
        fh.write('"line1"\n')
        fh.write('"line2"\n')
        fh.flush()
        yield fh
    os.remove('foo')

@fixture
def array_file():
    arr = np.arange(10)    
    with tempfile.NamedTemporaryFile(suffix='.gz') as tmp:
        np.savetxt(tmp.name, arr)
        yield tmp.name

def test_CacheCollection_load_existing_file(array_file): 
    c = CacheCollection('test')

    class Foo:
        name = 'foo'
        _file = array_file
        @c.cache
        async def generate(self):
            return np.array([])

    arr = asyncio.run(Foo().generate())
    np.testing.assert_array_equal(arr, np.arange(10))

def test_write_as_gz(file):
    write_as_gzip(file.name, 'foo.gz')
    line = gzip.open('foo.gz', 'rt').readline()
    assert line == '"line1"\n'

def test_write_as_json(file):
    write_as_json(file.name, 'foo.json', header=True)
    data = json.load(open('foo.json'))
    assert data[0] == 'line2'

def test_write_as_yaml(file):
    write_as_json(file.name, 'foo.yaml')
    data = yaml.load(open('foo.yaml'), yaml.loader.SafeLoader)
    assert data[0] == 'line1'

def test_CacheCollection_cache_array():
    c = CacheCollection('test')
    rand = np.random.randint(0, 100, 10)

    class Bar:
        name = 'bar'
        _file = None
        @c.cache
        async def generate(self):
            return rand

        async def write(self, arr):
            np.savetxt(self._file, arr, delimiter='\n')
            
    arr = asyncio.run(Bar().generate())
    np.testing.assert_array_equal(arr, rand)



import pytest
from pytest import fixture
import gzip
import asyncio
from genesynth.model import *

@fixture(scope='function')
def params():
    reseed(1)
    yield {
            'name': 'name',
            'size': 10,
    }

def test_integer(params):
    n = IntegerFixture(min=0, max=100, **params)
    #arr = np.array([37, 12, 72, 9, 75, 5, 79, 64, 16, 1]).astype(str)
    arr = np.array([41, 72, 0, 30, 14, 9, 18, 34, 39, 53]).astype(str)
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_serial(params):
    n = SerialFixture(min=0, step=2, **params)
    arr = np.array([0, 2, 4, 6, 8, 10, 12, 14, 16, 18]).astype(str)
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_boolean(params):
    n = BooleanFixture(**params)
    arr = np.array(['true', 'true', 'false', 'false', 'true', 'true', 'true', 'true', 'true', 'false'])
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_float(params):
    n = FloatFixture(min=0, max=10, **params)
    #arr = np.array([4.170220, 7.203245, 0.001143748, 3.023326, 1.467559, 0.9233859, 1.862602, 3.455607, 3.967675, 5.388167]).astype(str)
    arr = np.array(['4.17022004702574', '7.203244934421581', '0.0011437481734488664',
                '3.0233257263183977', '1.4675589081711304', '0.923385947687978',
                '1.862602113776709', '3.4556072704304777', '3.9676747423066994',
                '5.3881673400335695'])
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_decimal(params):
    n = DecimalFixture(min=0, max=10, precision=0, scale=5, **params)
    arr = np.array([4.17022, 7.20324, 0.00114, 3.02333, 1.46756, 0.92339, 1.8626, 3.45561, 3.96767, 5.38817]).astype(str)
    np.testing.assert_array_equal(asyncio.run(n.generate()), arr)

def test_map(params):
    children = {
        'id': SerialFixture(name='id', size=params['size'], min=0, step=1),
        'value': IntegerFixture(name='value', size=params['size'], min=0, max=10),
    }
    n = BaseMapFixture(children=children, **params)
    records = asyncio.run(n.generate())
    assert len(records) == len(children)
    assert set(records) == {'id', 'value'}
    np.testing.assert_array_equal(records['id'], np.arange(10).astype(str))

def test_array(params):
    children = [
        SerialFixture(name='id', size=params['size'], min=0, step=1),
        IntegerFixture(name='value', size=params['size'], min=0, max=10),
    ]
    n = BaseArrayFixture(children=children, **params)
    records = asyncio.run(n.generate())
    assert len(records) == len(children)
    np.testing.assert_array_equal(records[0], np.arange(10).astype(str))

def test_nest(params):
    fields = {
        'value': IntegerFixture(name='value', size=params['size'], min=0, max=10),
    }
    children = {
        'id': SerialFixture(name='id', size=params['size'], min=0, step=1),
        'json': BaseMapFixture(name='json', size=params['size'], children=fields),
    }
    n = BaseMapFixture(children=children, **params)
    records = asyncio.run(n.generate())
    assert set(records) == {'id', 'json'}
    np.testing.assert_array_equal(records['id'], np.arange(10).astype(str))
    assert set(records['json']) == {'value'}
    #np.testing.assert_array_equal(records['json']['value'], np.array([5, 8, 9, 5, 0, 0, 1, 7, 6, 9]).astype(str))
    np.testing.assert_array_equal(records['json']['value'], np.array([4, 6, 2, 8, 0, 6, 4, 5, 1, 1]).astype(str))

def test_write(params):
    fields = {
        'value': IntegerFixture(name='value', size=params['size'], min=0, max=10),
    }
    children = {
        'id': SerialFixture(name='id', size=params['size'], min=0, step=1),
        'json': JsonDataModel(name='json', size=params['size'], children=fields),
        'word': StringFixture(name='text', field='word', size=params['size'])
    }
    n = TableDataModel(children=children, **params)
    n.metadata = {'header': True}
    n.sep = b','
    asyncio.run(n.write())
    asyncio.run(n.save('foo'))
    with open('foo') as fh:
        assert fh.readline().rstrip('\n') == 'id,json,text'
        assert fh.readline().rstrip('\n') == '0,{"value": "4"},carol'
    os.remove('foo')

def test_nested_write(params):
    child_fields = {
        'value': IntegerFixture(name='value', size=params['size'], min=0, max=10),
    }
    fields = {
        'map': JsonDataModel(name='map', size=params['size'], children=child_fields),
    }
    children = {
        'id': SerialFixture(name='id', size=params['size'], min=0, step=1),
        'json': JsonDataModel(name='json', size=params['size'], children=fields),
        'word': StringFixture(name='text', field='word', size=params['size'])
    }
    n = TableDataModel(children=children, metadata={'sep': ',', 'header': True}, **params)
    n.metadata = {'header': True}
    n.sep = b','
    asyncio.run(n.write())
    asyncio.run(n.save('foo.gz'))
    with gzip.open('foo.gz', 'rt') as fh:
        assert fh.readline().rstrip('\n') == 'id,json,text'
        assert fh.readline().rstrip('\n') == '0,{"map": {"value": "4"}},carol'
    os.remove('foo.gz')



"""
IO is the utility library that manages how config and data gets loaded into the library as well as how it is written out.
"""

import os
import logging
import yaml
import gzip
import json
import functools
import yaml
import tempfile
import networkx as nx
import numpy as np
from mergedeep import merge, Strategy
from genesynth.utils import Hashabledict
from genesynth.extensions import datatypes
from genesynth.constraints import *

"""
1. construct graph based on input configuration
2. graph will be created recursively, and requires name to be unique at each level
3. data synthsis will only happen when type is delcared at lowest level within the tree
4. constraint is evaluated at each level with low level first, but overwritten by high level constraint
5. meta.parameter is configuration, constraint is validation
"""

logger = logging.getLogger(__name__)

def read_dot(filename):
    return nx.DiGraph(nx.nx_pydot.read_dot(filename))

def load_configs(*args):
    configs = []
    for filename in args:
        configs.append(load_config(filename))
    return merge(*configs, strategy=Strategy.REPLACE)

def load_config(filename):
    if filename.endswith('json'):
        with open(filename) as fh:
            data = json.load(fh)
    else:
        with open(filename) as fh:
            data = yaml.safe_load(fh)
    return data

def write_as_gzip(fh_obj, filename):
    with gzip.open(filename, 'wt') as fh:
        if isinstance(fh_obj, str):
            fh_obj = open(fh_obj)
        fh_obj.seek(0)
        for line in fh_obj:
            fh.write(line)

def write_as_json(fh_obj, filename, header=False):
    with open(filename, 'w') as fh:
        if isinstance(fh_obj, str):
            fh_obj = open(fh_obj)
        fh_obj.seek(0)
        if header:
            line = fh_obj.readline()
        records = [json.loads(line.rstrip('\n')) for line in fh_obj]
        json.dump(records, fh)

def write_as_yaml(fh_obj, filename, header=False):
    with open(filename, 'w') as fh:
        if isinstance(fh_obj, str):
            fh_obj = open(fh_obj)
        fh_obj.seek(0)
        if header:
            line = fh_obj.readline()
        records = [json.loads(line.rstrip('\n')) for line in fh_obj]
        yaml.dump(records, fh)

def unpack_constraints(constraints):
    items = []
    for constraint in constraints:
        if isinstance(constraint, dict):
            items.extend(list(zip(constraint.keys(), constraint.values())))
        else:
            items.append((constraint, None))
    return items

def schema_to_graph(G, fullname, params, size=0, root='root'):
    # TODO move this logic into io.py without disrupting package dependency
    # TODO incorporate Node and Relationship into node
    # TODO incorporate Constraint into node
    name = fullname.rsplit('.', 1)[-1]
    container = None
    type = params['type']
    if isinstance(type, (list, tuple)):
        type = type[0]
        container = 'array'
    elif type.startswith('[') and type.endswith(']'):
        type = type.strip('[]')
        container = 'array'
    metadata = params.get('metadata', {})
    metadata['size'] = metadata.get('size') or size
    metadata['sep'] = metadata.get('sep', '')
    foreign = metadata.pop('foreign', None)
    constraints = params.get('constraints', [])
    if constraints:
        constraints = unpack_constraints(constraints)
    if foreign:
        depends_on = f'{root}.{foreign["name"]}'
        node = datatypes['foreign'].from_params(name=fullname, graph=G, depends_on=depends_on, 
                                                    metadata=metadata, constraints=constraints, **metadata)
    elif container == 'array' and type == 'json':
        node = datatypes['json_array'].from_params(name=fullname, metadata=metadata, constraints=constraints, **metadata)
    elif container == 'array':
        node = datatypes['array'].from_params(name=fullname, metadata=metadata, constraints=constraints, **metadata)
    else:
        node = datatypes[type].from_params(name=fullname, metadata=metadata, constraints=constraints, **metadata)
    properties = params.get('properties')
    if properties is not None:
        children = {}
        for field, attributes in properties.items():
            field_fullname = f'{fullname}.{field}'
            child = schema_to_graph(G, field_fullname, attributes, size=size, root=root)
            children[child] = child
        if container == 'array' and type != 'json':
            node.children = tuple(children)
        else:
            node.children = Hashabledict(children)
        # add node to graph after setting data field children
        for child in children.values():
            G.add_edge(node, child) # add relationship type here
    G.add_node(node, label=fullname, xlabel=name, _id=fullname, type=type, metadata=metadata) # convert constraints into attributes
    return node

class CacheCollection:
    def __init__(self, name):
        self.name = name
        self.dir = tempfile.TemporaryDirectory(prefix=f'{name}_')

    def cache(self, fn):
        async def wrapped(obj, *args, **kwargs):
            if obj._file is None:
                arr = await fn(obj, *args, **kwargs)
                obj._file = os.path.join(self.dir.name, f'{obj.name}.gz')
                await obj.write(arr)
                return arr
            else:
                return np.loadtxt(obj._file)
        return wrapped



"""
Graph is solely responsible for entity relationships and logical order of operations.
Note that this is not intended to optimize processing order of operatsions and optimization.
The graph is required to be able to evaluate the dependency in isolation; this means no business logic.
"""

import enum
import networkx as nx
from genesynth import utils

class Node(enum.Enum):
    unique = 1 # node contains no duplicated value
    notnull = 2 # node contains no empty value
    sorted = 3 # ordered value, incremental if numeric, sorted if other
    subset = 4 # node contains value within the set (between is just (a...b) eps R)

class Relationship(enum.Enum):
    child = 1 # node is a child of parent
    subset = 2 # node contains subset of copy of parent
    ordered = 3 # node contains value that are based on ordered index of the parent, aka positional identity
    identity = 4 # node contains exact copy of parent
    incremental = 5 # node contains increasing value as parent

def find_node(G, name):
    for n in G.nodes:
        if n.name == name:
            return n

def find_child_node(G, parent, *children):
    node = find_node(G, parent)
    g = G.subgraph(nx.descendants(G, node))
    for child in children:
        parent = f'{parent}.{child}'
        node = find_node(g, parent) 
        g = G.subgraph(nx.descendants(g, node))
    return node

class Graph:
    """
    Handles all things graph traversal.
    Mainly used to detect deepest child (leaf) node and generate node processing sequence.
    """
    def __init__(self, G, name: str, ref: str = None, **attr):
        self.G = G
        self.name = name
        self.attr = attr
        self.ref = ref

    def subgraph(self, nodes):
        return self.G.subgraph(nodes)

    @property
    def nodes(self):
        return self.G.nodes

    @property
    def edges(self):
        return self.G.edges

    @property
    def root(self):
        return {n for n in self.G.nodes if self.G.in_degree(n) == 0}

    @property
    def model_nodes(self):
        return self.subgraph({n for n, d in self.G.nodes(data=True) if not n.is_data})

    @property
    def data_nodes(self):
        return self.subgraph({n for n, d in self.G.nodes(data=True) if n.is_data})

    def filter(self, **attrs):
        return self.subgraph({n for n, d in self.G.nodes(data=True) if d.items() >= attrs.items()})

    def find_node(self, name):
        return find_node(self.G, name)

    def find_child_node(self, parent, *children):
        return find_child_node(self.G, parent, *children)

    @property
    def leaf(self):
        return self.subgraph({n for n in self.G.nodes if self.G.in_degree(n) == 1 and self.G.out_degree(n) == 0})

    def parents(self, node):
        return self.subgraph(nx.ancestors(self.G, node))

    def children(self, node):
        return self.subgraph(nx.descendants(self.G, node))

    def node_degree(self):
        return dict(self.G.in_degree())

    def node_depth(self):
        depth = {}
        for root in self.root:
            depth.update(nx.shortest_path_length(self.G, root))
        return depth

    def degree_search(self):
        # TODO This needs to be replaced with depth
        sorted_degree = utils.sorted_groupby(self.node_depth().items(), lambda x: x[1], reverse=True)
        for degree, iterable in sorted_degree:
            yield degree, list(zip(*iterable))[0]

    #async def __aiter__(self):
    #    """
    #    degree_search -> iterate node -> dependency_graph -> generate
    #    """
    #    for degree, nodes in self.degree_search():
    #        for node in nodes:
    #            for n in self.parents(node).nodes:
    #                arr = await n.generate()
    #                yield n, arr
    #            arr = await node.generate()
    #            yield node, arr

    def __iter__(self):
        for degree, nodes in self.degree_search():
            for node in nodes:
                # TODO reverse order this iter
                for n in self.parents(node).nodes:
                    # TODO This needs to be fixed
                    yield n
                yield node

    async def generate(self):
        return set(self)
        #return {n: arr async for n, arr in self}

    def centrality(self):
        return nx.eigenvector_centrality(self.G)

    def transitive_closure(self):
        return nx.dag.transitive_closure(self.G)

    def transitive_reduction(self):
        return nx.dag.transitive_reduction(self.G)

    def domain_traversal(self, source=None):
        if source is None:
            source = list(self.G.nodes)[0]
        for node in nx.algorithms.traversal.bfs_edges(self.G, source):
            yield node

    def traversal(self, source=None):
        if source is None:
            source = list(self.G.nodes)[0]
        for node in nx.algorithms.traversal.dfs_edges(self.G, source):
            yield node

    def hydrated_traversal(self):
        pass

    def __del__(self, *args):
        del self.G



import os
from typing import List, Dict, Tuple, Any
from dataclasses import dataclass, field

import numpy as np
import openai
from genesynth.types import BaseMask
from genesynth.extensions import extensions

@extensions.register(['openai'])
@dataclass(unsafe_hash=True)
class BaseGenaiFixture(BaseMask):
    base_url = None
    env_var = 'OPENAI_API_KEY'

    prompt: str
    model: str = 'gpt-4o-mini'

    @property
    def api_key(self):
        return os.environ.get(self.env_var)

    async def generate(self):
        arr = []
        for _ in range(self.size):
            cli = openai.AsyncOpenAI(base_url=self.base_url, api_key=self.api_key)
            data = dict(role='user', content=self.prompt)
            stream = await cli.chat.completions.create(model=self.model, messages=[data], stream=True)
            text = ''.join([chunk.choices[0].delta.content or '' async for chunk in stream])
            arr.append(text.replace('\n', ' '))
        arr = np.array(arr)
        return self.apply_index(arr)


@extensions.register(['ollama'])
@dataclass(unsafe_hash=True)
class OllamaFixture(BaseGenaiFixture):
    base_url = 'http://localhost:11434/v1'

    prompt: str
    model: str = 'llama3.1:8b'

    @property
    def api_key(self):
        return 'fake_key'


@extensions.register(['google'])
@dataclass(unsafe_hash=True)
class GoogleFixture(BaseGenaiFixture):
    _base_url = 'https://{LOCATION}-aiplatform.googleapis.com/v1beta1/projects/{PROJECT}/locations/{LOCATION}/endpoints/openapi'

    prompt: str
    model: str = 'gemini-1.5-flash-001'

    @property
    def base_url(self):
        import google.auth as google_auth
        creds, project = google_auth.default()
        location = os.environ.get('LOCATION', 'us-central1')
        return self._base_url.format(PROJECT=project, LOCATION=location)

    @property
    def api_key(self):
        import google.auth
        import google.auth.transport.requests
        creds, project = google.auth.default()
        auth_req = google.auth.transport.requests.Request()
        creds.refresh(auth_req)
        return creds.token




import pytest
from pytest import fixture
import asyncio
from dataclasses import dataclass
import networkx as nx
import numpy as np
from genesynth.orchestration import *
from genesynth.worker import Runner
from genesynth.types import reseed, worker, SerialFixture, StringFixture

@fixture
def node():
    reseed(1)
    yield SerialFixture(name='serial', size=10)

@fixture
def string():
    reseed(1)
    yield StringFixture(name='text', field='word', size=10)

@fixture
def o():
    G = nx.DiGraph()
    runner = Runner(registry=worker)
    return Orchestration(G, runner=runner)

@pytest.mark.asyncio
async def test_process(o, node):
    arr = await o.process(node)
    np.testing.assert_array_equal(arr, np.arange(10).astype(str))

@pytest.mark.asyncio
async def test_worker_process(o, string):
    arr = await o.process(string)
    np.testing.assert_array_equal(arr, ['carol', 'square', 'returned', 'diary', 'lab', 'indicators', 'patterns', 'scenes', 'bi', 'alerts'])

@pytest.mark.asyncio
async def test_thread(o, node):
    arr = await o.thread(node)
    np.testing.assert_array_equal(arr, np.arange(10).astype(str))

@pytest.mark.asyncio
async def test_asyncio(o, node):
    arr = await o.asyncio(node)
    np.testing.assert_array_equal(arr, np.arange(10).astype(str))

def test_orchestration_read_config():
    o = Orchestration.read_config('tests/test.yaml')
    assert len(o.graph.nodes) == 7



import pytest
import os
from pytest import fixture
from genesynth.orchestration import *
from genesynth.cli import *

@fixture
def filename():
    reseed(1)
    return 'tests/test.yaml'

@fixture
def test_main(filename='tests/test.yaml'):
    main(filename, 'tests/out')
    assert os.path.isfile('tests/out')
    os.remove('tests/out')




import pytest
from pytest import fixture
import re
import asyncio
import psutil
from genesynth.worker import *

worker = WorkerRegistry()

class Foo:
    @worker.register
    def foo(self, a):
        return a

class Bar:
    @worker.register
    def foo(self, a):
        return a

    @worker.register
    async def bar(self, a):
        return a

    async def buzz(self, a):
        return a

@worker.register
def foo(a):
    return a

def test_registry():
    r = WorkerRegistry()
    class Bar:
        @r.register
        def bar(self, a):
            return a

    assert len(r) == 1

    b = Bar()
    assert b.bar(1) == 1
    assert Bar.bar.__qualname__ in r
    assert r[Bar.bar.__qualname__](b, 1) == 1

    f = Foo()
    assert f.foo(1) == 1
    assert Foo.foo.__qualname__ in worker
    assert worker[Foo.foo.__qualname__](f, 1) == 1

    assert foo(1) == 1
    assert foo.__qualname__ in worker
    assert worker[foo.__qualname__](1) == 1

@pytest.mark.asyncio
async def test_runner():
    workers = 2
    assert list(worker.keys()) == ['Foo.foo', 'Bar.foo', 'Bar.bar', 'foo']
    assert len(worker) == 4

    runner = Runner(registry=worker, workers=workers)
    assert runner.executor._max_workers == workers

    wraps = runner._wraps(Bar.bar)
    assert wraps.__qualname__ == Bar.bar.__qualname__
    assert wraps.__name__ == 'bar'

    assert not runner.executor._processes
    with pytest.raises(ValueError):
        assert 1 == await runner.run(Bar().foo, 1)
    assert len(runner.executor._processes) <= workers
    assert psutil.pid_exists(list(set(runner.executor._processes))[0])
    process_name = psutil.Process(list(set(runner.executor._processes))[0]).name().lower()
    assert re.search('python', process_name) or re.search('pytest', process_name)
    assert 1 == await runner.run(Bar().bar, 1)
    assert len(runner.executor._processes) <= workers
    assert 1 == await runner.run(Bar().buzz, 1)
    assert len(runner.executor._processes) <= workers




"""
Extensions is used for user defined types and allows the library to reference it during type conversion.
"""
import os
import glob
import importlib
from collections import ChainMap
from genesynth.types import types, Datatypes

extensions = Datatypes()  

datatypes = ChainMap(types, extensions)

for module in glob.glob(f'{os.path.dirname(__file__)}/*'):
    module, ext = os.path.splitext(os.path.basename(module))
    if ext == '.py':
        importlib.import_module(f'{__package__}.{module}')



"""
Model contains higher level abstraction that glues together underlying datatypes.
"""

import os
import shutil
import glob
import enum
import json
import asyncio
import tempfile
from typing import List, Dict, Tuple, Any
from dataclasses import dataclass, field
from contextlib import asynccontextmanager
import numpy as np
from scipy import stats
from genesynth.io import load_config, write_as_gzip
from genesynth.types import *
from genesynth.utils import iterate_lines

extensions = Datatypes()  

@dataclass(unsafe_hash=False)
class BaseDataModel(BaseMapFixture):
    """
    Handles data structure grouping based on the graph.
    Mainly deals with local root level structure hubs based on data type.
    Also handles aggregating cached resources and garbage collection.
    name: str - i.e. name of the table
    schema: str - i.e. namespace of the database such as schema
    sep: bytes - i.e. separator between columns, force conversion from string to bytes
    metadata: dict[str, str] - data model specific config (required) of the dataset
    constraints: list[str] - constrait to be added to the data, such as uniqueness check
    label: dict[str, str] - label to be added to the dataset that are not used for generation
    children: list - list od child dependencies
    """
    is_data = False
    full_key = False
    name: str
    schema: str = None # namespace from metadata
    sep: bytes = b''
    metadata: dict = field(default_factory=Hashabledict)
    constraints: tuple = field(default_factory=tuple)
    label: dict = field(default_factory=dict)
    children: dict = field(default_factory=dict)
    workload = WorkloadType.DEFAULT

    def __post_init__(self):
        super().__post_init__()
        self._dir = tempfile.TemporaryDirectory()
        self.metadata = Hashabledict(self.metadata)
        self.constraints = tuple(self.constraints)
        self.label = Hashabledict(self.label)
        self.children = Hashabledict(self.children)

    async def __aiter__(self):
        pass

    async def generate(self):
        gens = [asyncio.create_task(n.generate()) for n in self.children.values()]
        done, pending = await asyncio.wait(gens)
        return [await f for f in done]

    @asynccontextmanager
    async def _filename(self, path=None):
        if path is None:
            filename = os.path.join(self._dir.name, self.name)
        else:
            filename = os.path.join(path, self.name)
        with open(filename, 'wb+') as fh:
            yield fh

    async def merge(self, nodes, path=None):
        """Combine children cache files into one
        """
        filenames = [node._file for node in nodes]
        length = 0
        async with self._filename(path) as fh:
            for lines in iterate_lines(*filenames):
                fh.write(self.sep.join(lines))
                fh.write(b'\n')
                await asyncio.sleep(0)
                length += 1
        assert length == self.size or length - 1 == self.size, f'expected {self.size} data row, got {length}'

    async def write(self):
        if self._file is not None:
            return
        gens = []
        for n in self.children.values():
            gens.append(n)
            n._path = self._dir.name
            # TODO defer this to orchestrator
            await n.write()
        await self.merge(gens, path=self._dir.name)
        self._file = os.path.join(self._dir.name, self.name)

    async def save(self, filename=None):
        if self._file is None:
            await self.write()
        if filename is not None:
            _, ext = os.path.splitext(filename)
            if ext == '.gz':
                write_as_gzip(self._file, filename)
            else:
                shutil.copy(self._file, filename)

    def __hash__(self):
        return hash(str(self))


@types.register(['object', 'table'])
@dataclass(unsafe_hash=False)
class TableDataModel(BaseDataModel):
    # TODO change this to have default keys
    metadata: dict = field(default_factory=dict)

    def __post_init__(self):
        super().__post_init__()
        # TODO allow support for string instead of just bytes
        self.sep = self.metadata.get('sep', self.sep)
        if isinstance(self.sep, str):
            self.sep = self.sep.encode('utf-8')

    @property
    def has_header(self):
        return self.metadata.get('header')

    @property
    def has_footer(self):
        return self.metadata.get('footer')

    @property
    def header(self):
        if isinstance(self.metadata['header'], bool):
            return self.sep.join(n.name.encode('utf-8') for n in self.children.values())
        return self.metadata['header'].encode('utf-8')

    @property
    def footer(self):
        return self.metadata.get('footer', '').encode('utf-8')

    @asynccontextmanager
    async def _filename(self, path=None):
        """context mamager to create the staged temporary output file for the data model.
        """
        if path is None:
            filename = os.path.join(self._dir.name, self.name)
        else:
            filename = os.path.join(path, self.name)
        with open(filename, 'wb+') as fh:
            if self.has_header:
                fh.write(self.header)
                fh.write(b'\n')
            yield fh
            if self.has_footer:
                fh.write(self.footer)

    def __hash__(self):
        return hash(str(self))


@types.register(['json'])
@dataclass(unsafe_hash=True)
class JsonDataModel(BaseDataModel):
    is_data = True
    async def merge(self, nodes, path=None):
        filenames = {node.name: node._file for node in nodes}
        async with self._filename(path) as fh:
            for lines in iterate_lines(*filenames.values()):
                # TODO add support for array
                record = {key if self.full_key else key.split('.')[-1] : \
                            json.loads(value) if isinstance(node, BaseMapFixture) else value 
                            for node, key, value in zip(nodes, filenames.keys(), lines)}
                fh.write(json.dumps(record, default=lambda x: x.decode()).encode('utf-8'))
                fh.write(b'\n')
                await asyncio.sleep(0)
            fh.seek(0)
            length = len(list(fh))
        assert length == self.size, f'expected {self.name} to have {self.size} data row, got {length}'

    async def write(self):
        gens = []
        if self._path is None:
            path = self._dir.name
        else:
            path = os.path.join(self._path, self.name)
        if not os.path.isdir(path):
            os.mkdir(path)
        for n in self.children.values():
            gens.append(n)
            n._path = path
            # TODO defer this to orchestrator
            await n.write()
        await self.merge(gens, path)
        self._file = os.path.join(path, self.name)

@types.register(['json_array'])
@dataclass(unsafe_hash=True)
class JsonArrayDataModel(JsonDataModel):
    is_data = True
    async def merge(self, nodes, path=None):
        filenames = {node.name: node._file for node in nodes}
        async with self._filename(path) as fh:
            for lines in iterate_lines(*filenames.values()):
                # TODO add support for array
                record = {key if self.full_key else key.split('.')[-1] : json.loads(value) if isinstance(node, BaseMapFixture) else value 
                            for node, key, value in zip(nodes, filenames.keys(), lines)}
                fh.write(json.dumps([record], default=lambda x: x.decode('ascii')).encode('utf-8'))
                fh.write(b'\n')
                await asyncio.sleep(0)
            fh.seek(0)
            length = len(list(fh))
        assert length == self.size, f'expected {self.size} data row, got {length}'



#!/usr/bin/env python
import sys
import os
import argparse
import asyncio
from genesynth.orchestration import *
from genesynth.utils import clean_text

parser = argparse.ArgumentParser()
parser.add_argument('-f', '--filename', required=True, nargs='+', help='schema file')
parser.add_argument('-o', '--output', help='output filename if given. defaults to input filename without extension')
parser.add_argument('--string', action='store_true', help='keep all type as string')
parser.add_argument('--stdout', action='store_true', help='DEPRECATED print output to stdout')

def main(filename, output=None, string=False):
    pipe = Orchestration.read_config(*filename)
    pipe.run()
    if output is None:
        asyncio.run(pipe.root.save())
        with open(pipe.root._file) as fh:
            for line in fh:
                if string:
                    sys.stdout.write(line)
                else:
                    sys.stdout.write(clean_text(line))
    else:
        asyncio.run(pipe.root.save(output))

if __name__ == '__main__':
    args = parser.parse_args()
    main(args.filename, args.output, args.string)



#!/usr/bin/env python
import os
import importlib
import setuptools

name = 'genesynth'
version = importlib.import_module(f'{name}.version').__version__
description = f'Python package: {name}'

def readme(filename='README.md'):
    try:
        with open(filename) as fh:
            return fh.read()
    except:
        pass

def dep(filename='requirements.txt'):
    try:
        with open(filename) as fh:
            return [line.rstrip('\n') for line in fh]
    except:
        return []

if __name__ == '__main__':
    setuptools.setup(
        name=name,
        version=version,
        license='MIT',
        author='Gang Huang',
        description=description,
        long_description=readme(),
        long_description_content_type="text/x-rst",
        packages=setuptools.find_packages(),
        python_requires='>=3.9',
        install_requires=dep(),
    )



"""
Mat contains all matrix and masking operations that meets the constraint needs.
"""

import logging
import enum
import numpy as np
from scipy import stats

class StatsModel(enum.Enum):
    uniform = stats.uniform
    normal = stats.norm
    student_t = stats.t
    binomial = stats.binom
    poisson = stats.poisson
    alpha = stats.alpha
    beta = stats.beta
    gamma = stats.gamma
    cosine = stats.cosine
    fisher = stats.f
    power_law = stats.powerlaw
    bootstrap = stats.bootstrap

def identity(size: int, matmul=False):
    """return index array that will produce identical array
    when applied to the original array
    if matmul is true, it will return boolean array
        otherwise it will return index array
    """
    if matmul:
        return np.ones(size).astype(bool)
    return np.arange(size)

def unique(arr: np.array):
    """return subarray of unique values
    """
    return np.unique(arr)

def sample(size: int, options, replace=True):
    """return sampled value from options
    if options is a dict, sampling probability distribution will be 
        generated from normalizing the value from options, otherwise 
        it will be a uniform distribution iterated over options
    """
    if isinstance(options, dict):
        p = np.array(list(options.values()))
        p = p / p.sum()
        options = list(options)
    else:
        p = None
    return np.random.choice(options, size, replace=replace, p=p)

def mask(arr: np.array, low, high=None):
    """return index of input value against the array
    length of the array depends on number matches
    """
    if high is not None:
        return np.where((arr>=low) & (arr<high))[0]
    else:
        return np.where(arr==low)[0]

def mask_index(arr: np.array, low, high=None):
    """return masking index for value that matches
    """
    if high is not None:
        return np.where((arr>=low) & (arr<high), True, False)
    else:
        return np.where(arr==low, True, False)

def null_percent(size: int, percent: float = 0.):
    """return masking array for null value drawn from a binomial distribution.
    """
    return np.random.binomial(1, percent/100., size) == 1

def nullable(arr, percent=0, mask=None, null=''):
    """return masked array with null being the fill_value
    make sure that null value matches the datatype of array, or it will error
    use arr.filled() to return array with mask filled
    """
    if mask is None:
        return np.ma.MaskedArray(arr, mask=null_percent(arr.size, percent), fill_value=null)
    else:
        return np.ma.MaskedArray(arr, mask=mask, fill_value=null)

def ordered_index(arr: np.array):
    """return index to sort array in incremental order
    """
    return np.argsort(arr, kind='stable')

def indexed(arr: np.array, idx: np.array):
    """return array with index applied
    """
    return arr[idx]

def revert_ordered_index(arr: np.array):
    """return index to unsort array
    """
    return np.argsort(ordered_index(arr), kind='stable')

def cosine_similarity(a: np.array, b: np.array):
    """return cosine similarity between two arrays
    """
    return np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))

def stats_model(*args, model: str = 'uniform', **kwargs):
    """return stats model based on input parameters
       model supported are:
            alpha
            beta
            binomial
            cosine
            fisher
            gamma
            normal
            poisson
            power_law
            student_t
            uniform
       parameters to the model can be passed in as ordered or key-word wildcards 
    """
    if isinstance(model, str):
        return StatsModel[model].value(*args, **kwargs)
    else:
        return StatsModel(model).value(*args, **kwargs)

def stats_model_generate(size: int, min: float, max: float, *args, model: str = 'uniform', unique=False, **kwargs):
    """return array based on the statistical distribution
       parameters to the model can be passed in as ordered or key-word wildcards 
    """
    if model == 'uniform':
        m = stats_model(min, max, *args, model=model)
    else:
        m = stats_model(*args, model=model, **kwargs)
    p99_low, p99_high = m.interval(0.99)
    p_low, p_high = m.cdf(min), m.cdf(max)
    if unique:
        p = np.linspace(p_low, p_high, size)
    else:
        p = np.random.rand(size) * (p_high - p_low) + p_low
    return m.ppf(p)

def stats_model_fit(arr: np.array, model: str = 'uniform'):
    """return model best fit parameters based on input array
       model supported are:
            alpha
            beta
            cosine
            fisher
            gamma
            normal
            power_law
            student_t
            uniform
       parameters to the model can be passed in as ordered or key-word wildcards 
    """
    return StatsModel[model].value.fit(arr)



"""
Core library that determines how data will be generated based on respective types (dataclass).
It will be possible to represent each type as a protocol buffer for serialization.
All type must provide async generate method and optional write method.
"""

import os
import enum
import random
import hashlib
import json
from typing import List, Dict, Tuple, Any
from dataclasses import dataclass, field
from datetime import datetime, date, time
from functools import lru_cache
import httpx
import numpy as np
import pandas as pd
from scipy import stats
from mimesis import Generic
from mimesis.random import Random
from mimesis.locales import Locale
from mimesis.builtins import USASpecProvider
from genesynth.worker import WorkloadType, Runner, WorkerRegistry
from genesynth.constraints import *
from genesynth.graph import nx, find_node, find_child_node
from genesynth.utils import Hashabledict, sorted_groupby
from genesynth import mat

def reseed(seed=None):
    BaseMask.seed = seed
    np.random.seed(seed)
    random.seed(seed)

worker = WorkerRegistry()

class Datatypes(dict):
    def register(self, alias: List):
        def wraps(obj):
            assert BaseMask in obj.__mro__, f'type {obj.__name__} needs to be subclass of BaseMask'
            for name in alias:
                self[name] = obj
            return obj
        return wraps

types = Datatypes()
datatypes = types

@dataclass(unsafe_hash=True)
class BaseMask:
    is_data = True
    seed = None
    name: str
    size: int
    workload = WorkloadType.IO
    null = None
    _metadata = {}
    _constraints = []
    _index = None
    _mask = None
    _file = None # cache filename
    _hashfile = True
    _path = None # cache directory
    _defer = None # parent node if deferred
    _dist = None

    # TODO handle notnull and unique constraits

    @classmethod
    def from_params(cls, metadata={}, constraints=[], **kwargs):
        # TODO add type conversion based on constraints
        fields = set(cls.__dataclass_fields__.keys())
        keys = set(kwargs.keys())
        params = {field: kwargs[field] for field in fields & keys}
        obj = cls(**params)
        obj.metadata = Hashabledict(metadata)
        if 'dist' in metadata:
            obj._dist = metadata.pop('dist')
        obj.constraints = tuple(cls.unpack_constraints(constraints))
        return obj

    @property
    def unique(self):
        return 'unique' in self._constraints

    @property
    def dist(self):
        if self._dist:
            return next(iter(self._dist.keys()))
        else:
            return 'uniform'

    @property
    def dist_params(self):
        if self._dist:
            return next(iter(self._dist.values()))
        else:
            return {}

    @staticmethod
    def unpack_constraints(constraints):
        return {k: list(filter(None, (i[-1] for i in v))) 
            for k, v in sorted_groupby(constraints, lambda x: x[0])}

    @lru_cache(maxsize=None)
    def mask(self):
        if 'notnull' in self._constraints:
            return mat.null_percent(self.size, percent=0)
        elif 'nullable' in self._constraints:
            mask = mat.identity(self.size, matmul=True)
            for value in self._constraints['nullable']:
                mask &= mat.null_percent(self.size, percent=value)
            return mask

    def dist_generate(self, *args, **kwargs):
        if not args and self.dist == 'uniform':
            args = (0, 1)
        return mat.stats_model_generate(self.size, *args, model=self.dist, unique=self.unique, **self.dist_params)

    @staticmethod
    def dist_fit(arr):
        return mat.stats_model_fit(arr, model=self.dist, **self.dist_params)

    @staticmethod
    def index_value(arr):
        return mat.ordered_index(arr)

    @staticmethod
    def subset_index(arr, subset):
        index = mat.identity(arr.size)
        mask = mat.identity(self.size, matmul=True) == False
        for unique in np.unique(subset):
            mask |= mat.mask_index(arr, unique)
        return index[mask]

    def index(self, arr):
        index = mat.identity(self.size)
        if 'sorted' in self._constraints:
            index = mat.ordered_index(arr)
        if self.unique:
            assert len(index) == np.unique(index).size, f'{self.name} should be unique, found dupe in index'
            unique = np.unique(arr)
            assert len(index) == unique.size, f'{self.name} only has {unique.size}/{len(index)} unique value'
        return index

    def apply_index(self, arr, null=''):
        mask = self.mask()
        index = self.index(arr)
        # TODO remove the need to cast array as str
        return mat.nullable(arr[index].astype(str), mask=mask, null=null).filled()

    async def generate(self):
        raise NotImplementedError(f'{self.__class__.__name__} does not have generate defined')

    async def write(self):
        if self._file is not None:
            return
        if isinstance(self, BaseArrayFixture):
            arr = await self.generate()
            arr = np.array(arr)
            if arr.ndim > 2:
                arr = np.squeeze(arr) # TODO temporary hack. need to figure out how to output this better
        else:
            arr = await self.generate()
        if self._path is not None:
            if self._hashfile:
                filename = os.path.join(self._path, hashlib.md5(self.name.encode('ascii')).hexdigest())
            else:
                filename = os.path.join(self._path, self.name)
        else:
            if self._hashfile:
                filename = hashlib.md5(self.name.encode('ascii')).hexdigest()
            else:
                filename = self.name
        np.savetxt(filename, arr, fmt='%s', delimiter='\n')
        self._file = filename

    def __str__(self):
        return f"{self.__class__.__name__}(name='{self.name}', size={self.size})"

    def __del__(self, *args, **kwargs):
        if self._file and os.path.isfile(self._file):
            os.remove(self._file)

@dataclass(unsafe_hash=True)
class BaseNumberFixture(BaseMask):
    null = np.nan
    pass

@types.register(['enum'])
@dataclass(unsafe_hash=True)
class BaseTextFixture(BaseMask):
    options: tuple = field(default_factory=tuple)
    replace: bool = True

    def __post_init__(self):
        self.options = tuple(self.options)
        if self.unique:
            self.replace = False

    async def generate(self):
        arr = mat.sample(self.size, self.options, replace=self.replace)
        return self.apply_index(arr)

@types.register(['text'])
@dataclass(unsafe_hash=True)
class BaseTextFixture(BaseMask):
    length: int = None

    def __post_init__(self):
        self.random = Random(self.seed)

    async def generate(self):
        # TODO add support to ensure unique constraint
        arr = np.array([self.random.randstr()[:self.length] for _ in range(self.size)])
        return self.apply_index(arr)

@types.register(['timestamp', 'datetime'])
@dataclass(unsafe_hash=True)
class BaseTimestamp(BaseMask):
    min: datetime = datetime.fromtimestamp(0)
    max: datetime = datetime.now()
    posix: bool = False

    async def generate(self):
        arr = pd.date_range(self.min, self.max, periods=self.size)
        if not self.posix:
            arr = arr.to_pydatetime()
        else:
            arr = (arr.values.astype(int) // 10**9) + (arr.microsecond / 10**6)
        return self.apply_index(arr)

@types.register(['date'])
@dataclass(unsafe_hash=True)
class BaseDate(BaseTimestamp):
    async def generate(self):
        arr = pd.date_range(self.min, self.max, periods=self.size).date
        return self.apply_index(arr)

@types.register(['time'])
@dataclass(unsafe_hash=True)
class BaseTime(BaseTimestamp):
    min: time = time(0)
    max: time = time(23, 59, 59)

    async def generate(self):
        min = datetime.fromtimestamp(0).replace(hour=self.min.hour, minute=self.min.minute,
            second=self.min.second, microsecond=self.min.microsecond)
        max = datetime.fromtimestamp(0).replace(hour=self.max.hour, minute=self.max.minute,
            second=self.max.second, microsecond=self.max.microsecond)
        arr = pd.date_range(min, max, periods=self.size).time
        return self.apply_index(arr)

@types.register(['foreign'])
@dataclass(unsafe_hash=True)
class BaseForeign(BaseMask):
    depends_on: str
    graph: nx.DiGraph
    _node = None

    @property
    def node(self):
        if self._node is None:
            parent, *child = self.depends_on.split('.')
            self._node = find_child_node(self.graph, parent, *child)
            self._node._path = self._path
            self.resolve_fkey_edge()
        return self._node

    def resolve_fkey_edge(self):
        if (self.node, self) not in self.graph.edges:
            self.graph.add_edge(self.node, self, label='fkey', type='fkey')

    async def generate(self):
        arr = await self.node.generate()
        return self.apply_index(arr)

@types.register(['array', 'list', 'tuple'])
@dataclass(unsafe_hash=True)
class BaseArrayFixture(BaseMask):
    null = []
    children: tuple = field(default_factory=tuple)

    def __post_init__(self):
        self.children = tuple(self.children)

    async def __aiter__(self):
        for node in self.children:
            if isinstance(node, BaseMask):
                node.size = self.size
                yield await node.generate()
            else:
                yield node

    async def generate(self):
        return [arr async for arr in self]

@types.register(['map', 'struct'])
@dataclass(unsafe_hash=True)
class BaseMapFixture(BaseMask):
    null = {}
    children: dict = field(default_factory=dict)

    def __post_init__(self):
        self.children = Hashabledict(self.children)

    async def __aiter__(self):
        for key, node in self.children.items():
            if isinstance(node, BaseMask):
                node.size = self.size
                arr = await node.generate()
                yield key, arr
            else:
                yield node

    async def generate(self):
        return {key: arr async for key, arr in self}

@types.register(['integer'])
@dataclass(unsafe_hash=True)
class IntegerFixture(BaseNumberFixture):
    min: int = 0
    max: int = 100
    null = None

    # TODO add type conversio to Serial when constraint is incremental

    async def generate(self):
        arr = mat.stats_model_generate(self.size, self.min, self.max, model=self.dist, unique=self.unique, **self.dist_params).astype(int)
        return self.apply_index(arr)

@types.register(['serial'])
@dataclass(unsafe_hash=True)
class SerialFixture(BaseNumberFixture):
    min: int = 0
    step: int = 1
    null = 0

    async def generate(self):
        arr = np.arange(self.min, self.min + self.size * self.step, self.step)
        return self.apply_index(arr)

@types.register(['boolean'])
@dataclass(unsafe_hash=True)
class BooleanFixture(BaseNumberFixture):
    async def generate(self):
        arr = np.random.randint(0, 2, self.size).astype(bool)
        arr = np.where(arr==True, 'true', np.where(arr==False, 'false', arr))
        return self.apply_index(arr)

@types.register(['float', 'double'])
@dataclass(unsafe_hash=True)
class FloatFixture(BaseNumberFixture):
    min: int = -1
    max: int = 1

    async def generate(self):
        arr = mat.stats_model_generate(self.size, self.min, self.max, model=self.dist, unique=self.unique, **self.dist_params)
        return self.apply_index(arr)

@types.register(['decimal', 'numeric'])
@dataclass(unsafe_hash=True)
class DecimalFixture(FloatFixture):
    precision: int = 16
    scale: int = 3

    async def generate(self):
        arr = mat.stats_model_generate(self.size, self.min, self.max, model=self.dist, unique=self.unique, **self.dist_params)
        arr = np.round(arr, self.scale)
        return self.apply_index(arr)

@types.register(['string'])
@dataclass(unsafe_hash=True)
class StringFixture(BaseTextFixture):
    null = ''
    subtype: str = 'text'
    field: str = 'sentence'
    locale = Locale.EN

    def __post_init__(self):
        self.generic = Generic(locale=self.locale, seed=self.seed)

    @worker.register
    async def generate(self):
        func = getattr(self.generic, self.subtype)
        if self.field is not None:
            func = getattr(func, self.field)
        # TODO add support to ensure unique constraint
        arr = np.array([str(func())[:self.length] for _ in range(self.size)])
        return self.apply_index(arr)

    @worker.register
    async def write(self):
        return await super().write()

@types.register(['password'])
@dataclass(unsafe_hash=True)
class BcryptPassword(BaseTextFixture):
    rounds: int = 12
    prefix: str = '2b'

    async def generate(self):
        arr = np.array([f'${self.prefix}${self.rounds}${self.random.randstr(length=53)}' for _ in range(self.size)])
        return self.apply_index(arr)

